import unittest

import oaas
import oaas.registry

from oaas_simple.client.client import OaasSimpleClient
from oaas_simple.server.server import OaasSimpleServer


@oaas.service("test-service")
class TestCallService:
    def echo_data(self, *, data: str) -> str:
        return data


@oaas.client("test-service")
class TestCallClient:
    def echo_data(self, *, data: str) -> str:
        ...


oaas.register_server_provider(OaasSimpleServer())
oaas.register_client_provider(OaasSimpleClient())

oaas.serve()


class TestGrpcSerialization(unittest.TestCase):
    def test_serialization(self) -> None:
        client = oaas.get_client(TestCallClient)
        self.assertEqual("abc", client.echo_data(data="abc"))



from typing import List

import netifaces
import oaas.registry


def find_ips() -> List[oaas.registry.IpLink]:
    result: List[oaas.registry.IpLink] = []

    for interface in netifaces.interfaces():
        ifaddresses = netifaces.ifaddresses(interface)

        if netifaces.AF_INET not in ifaddresses:
            continue

        for link in ifaddresses[netifaces.AF_INET]:
            result.append(link["addr"])

    return result



import os
import re

from oaas.registry import ServiceAddress

IP_PORT_ADDRESS = re.compile("^(.+)(:\d+)?$")


def find_registry() -> ServiceAddress:
    if "OAAS_REGISTRY" in os.environ:
        return read_oaas_registry_from_environ()

    return create_static_search_list()


def read_oaas_registry_from_environ() -> ServiceAddress:
    oaas_registry = os.environ["OAAS_REGISTRY"]
    m = IP_PORT_ADDRESS.match(oaas_registry)

    if not m:
        raise Exception(
            f"Unable to parse OAAS_REGISTRY " f"environment variable: {oaas_registry}"
        )

    return {"port": int(m.group(2)), "addresses": [m.group(1)]}


def create_static_search_list() -> ServiceAddress:
    return {
        "port": 8999,
        "addresses": [
            "localhost",
            # docker
            "172.17.0.1",
            # kubernetes
            "oaas-registry",
            "oaas-registry.oaas-registry.svc.cluster.local",
        ],
    }



from concurrent import futures

import grpc
import oaas
import oaas._registrations as registrations

from oaas_simple.registry import oaas_registry
from oaas_simple.rpc import call_pb2_grpc
from oaas_simple.server.find_ips import find_ips
from oaas_simple.server.service_invoker_proxy import ServiceInvokerProxy


class OaasSimpleServer(oaas.ServerMiddleware):
    def __init__(self, *, port=8999):
        super(OaasSimpleServer, self).__init__()
        self.port = port

    def serve(self) -> None:
        self.server = self.start_server()
        self.register_services_into_registry()

    def join(self) -> None:
        self.server.wait_for_termination()

    def start_server(self):
        server_address: str = f"[::]:{self.port}"
        server = grpc.server(futures.ThreadPoolExecutor())
        call_pb2_grpc.add_ServiceInvokerServicer_to_server(
            ServiceInvokerProxy(), server
        )
        port = server.add_insecure_port(server_address)
        server.start()
        return server

    def register_services_into_registry(self):
        for service_definition, _ in registrations.services.items():
            if service_definition.name == "oaas-registry":
                continue

            oaas_registry().register_service(
                {
                    "name": service_definition.name,
                    "protocol": "simple",
                },
                {
                    "port": self.port,
                    "addresses": find_ips(),
                },
            )



from typing import Any

from oaas import ClientDefinition

from oaas_simple.data import create_data, from_data
from oaas_simple.rpc import call_pb2


class SingleMethod:
    def __init__(
        self, *, stub: Any, client_definition: ClientDefinition, method_name: str
    ):
        self.stub = stub
        self.client_definition = client_definition
        self.method = method_name

    def __call__(self, *args, **kwargs):
        parameters = []

        for arg in args:
            param = call_pb2.ServiceCallParam(name=None, data=create_data(arg))
            parameters.append(param)

        for arg_name, arg in kwargs.items():
            param = call_pb2.ServiceCallParam(name=arg_name, data=create_data(arg))
            parameters.append(param)

        response: call_pb2.Data = self.stub.InvokeMethod(
            call_pb2.ServiceCall(  # ignore: type
                namespace=None,
                service=self.client_definition.name,
                version=None,
                method=self.method,
                parameters=parameters,
            )
        )

        return from_data(response)



from typing import Any, Dict

import grpc
import oaas
import socket
import random

from oaas import ClientDefinition
from oaas.registry import ServiceAddress

from oaas_simple.client import registry_discovery
from oaas_simple.client.service_client_proxy import ServiceClientProxy
from oaas_simple.registry import oaas_registry


def is_someone_listening(host_address: str, port: int):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as a_socket:
        location = (host_address, port)
        try:
            result_of_check = a_socket.connect_ex(location)

            return result_of_check == 0
        except Exception:
            return False


class OaasSimpleClient(oaas.ClientMiddleware):
    def __init__(self) -> None:
        self._client_to_address: Dict[str, str] = dict()
        self._channels: Dict[str, Any] = dict()

    def create_client(self, cd: ClientDefinition) -> Any:
        if cd.name == "oaas-registry":
            service_address = registry_discovery.find_registry()
        else:
            service_address = oaas_registry().resolve_service(
                {
                    "name": cd.name,
                    "protocol": "simple",
                    "namespace": "",
                    "version": "",
                }
            )

        address = self._client_to_address.get(cd.name, None)

        if address:
            return self._channels[address]

        channel = self.create_channel(cd, service_address)

        return ServiceClientProxy(client_definition=cd, channel=channel)

    def create_channel(self, cd: ClientDefinition, service_address: ServiceAddress):
        port = service_address["port"]
        addresses = service_address["addresses"]
        random.shuffle(addresses)

        for host_address in addresses:
            address = f"{host_address}:{port}"

            # if we already have a channel for this address, we don't need
            # to check if the port is open, and establish the connection,
            # but simply reuse the same channel.
            if address in self._channels:
                self._client_to_address[cd.name] = address
                return self._channels[address]

            if not is_someone_listening(host_address, port):
                continue

            channel = grpc.insecure_channel(address)
            self._channels[address] = channel
            self._client_to_address[cd.name] = address

            return channel

        raise Exception(
            f"Unable to find port {port} open on any of "
            f"the {service_address['addresses']}"
        )

    def can_handle(self, cd: ClientDefinition) -> bool:
        # FIXME: should at least check the registered types
        return True



from typing import Optional

import oaas
import oaas.registry


@oaas.client("oaas-registry")
class OaasRegistry(oaas.registry.OaasRegistry):
    ...


_oaas_registry: Optional[OaasRegistry] = None


def oaas_registry() -> OaasRegistry:
    global _oaas_registry

    if _oaas_registry:
        return _oaas_registry

    _oaas_registry = oaas.get_client(OaasRegistry)  # type: ignore

    return _oaas_registry



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: call.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="call.proto",
    package="gr",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\ncall.proto\x12\x02gr"*\n\x04\x44\x61ta\x12\t\n\x01s\x18\x01 \x01(\t\x12\t\n\x01\x62\x18\x02 \x01(\x0c\x12\x0c\n\x04json\x18\x03 \x01(\t"8\n\x10ServiceCallParam\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x16\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\x08.gr.Data"|\n\x0bServiceCall\x12\x11\n\tnamespace\x18\x01 \x01(\t\x12\x0f\n\x07service\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x0e\n\x06method\x18\x04 \x01(\t\x12(\n\nparameters\x18\x05 \x03(\x0b\x32\x14.gr.ServiceCallParam2=\n\x0eServiceInvoker\x12+\n\x0cInvokeMethod\x12\x0f.gr.ServiceCall\x1a\x08.gr.Data"\x00\x62\x06proto3',
)


_DATA = _descriptor.Descriptor(
    name="Data",
    full_name="gr.Data",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="s",
            full_name="gr.Data.s",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="b",
            full_name="gr.Data.b",
            index=1,
            number=2,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="json",
            full_name="gr.Data.json",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=18,
    serialized_end=60,
)


_SERVICECALLPARAM = _descriptor.Descriptor(
    name="ServiceCallParam",
    full_name="gr.ServiceCallParam",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="name",
            full_name="gr.ServiceCallParam.name",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="gr.ServiceCallParam.data",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=62,
    serialized_end=118,
)


_SERVICECALL = _descriptor.Descriptor(
    name="ServiceCall",
    full_name="gr.ServiceCall",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="namespace",
            full_name="gr.ServiceCall.namespace",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="service",
            full_name="gr.ServiceCall.service",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="version",
            full_name="gr.ServiceCall.version",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="method",
            full_name="gr.ServiceCall.method",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parameters",
            full_name="gr.ServiceCall.parameters",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=120,
    serialized_end=244,
)

_SERVICECALLPARAM.fields_by_name["data"].message_type = _DATA
_SERVICECALL.fields_by_name["parameters"].message_type = _SERVICECALLPARAM
DESCRIPTOR.message_types_by_name["Data"] = _DATA
DESCRIPTOR.message_types_by_name["ServiceCallParam"] = _SERVICECALLPARAM
DESCRIPTOR.message_types_by_name["ServiceCall"] = _SERVICECALL
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Data = _reflection.GeneratedProtocolMessageType(
    "Data",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATA,
        "__module__": "call_pb2"
        # @@protoc_insertion_point(class_scope:gr.Data)
    },
)
_sym_db.RegisterMessage(Data)

ServiceCallParam = _reflection.GeneratedProtocolMessageType(
    "ServiceCallParam",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVICECALLPARAM,
        "__module__": "call_pb2"
        # @@protoc_insertion_point(class_scope:gr.ServiceCallParam)
    },
)
_sym_db.RegisterMessage(ServiceCallParam)

ServiceCall = _reflection.GeneratedProtocolMessageType(
    "ServiceCall",
    (_message.Message,),
    {
        "DESCRIPTOR": _SERVICECALL,
        "__module__": "call_pb2"
        # @@protoc_insertion_point(class_scope:gr.ServiceCall)
    },
)
_sym_db.RegisterMessage(ServiceCall)


_SERVICEINVOKER = _descriptor.ServiceDescriptor(
    name="ServiceInvoker",
    full_name="gr.ServiceInvoker",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=246,
    serialized_end=307,
    methods=[
        _descriptor.MethodDescriptor(
            name="InvokeMethod",
            full_name="gr.ServiceInvoker.InvokeMethod",
            index=0,
            containing_service=None,
            input_type=_SERVICECALL,
            output_type=_DATA,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_SERVICEINVOKER)

DESCRIPTOR.services_by_name["ServiceInvoker"] = _SERVICEINVOKER

# @@protoc_insertion_point(module_scope)
